/////////////////////////////////////////////////////////////////////////////
// File:            gnDNXSource.h
// Purpose:         Implements gnBaseSource for .DNX files
// Description:     
// Changes:        
// Version:         libGenome 0.5.1 
// Author:          Aaron Darling 
// Modified by:     
// Copyright:       (c) Aaron Darling 
// Licenses:        See COPYING file for details
/////////////////////////////////////////////////////////////////////////////
#include "gn/gnDNXSource.h"
#include "gn/gnSourceSpec.h"
#include "gn/gnStringSpec.h"
#include "gn/gnSourceFactory.h"
#include "gn/gnFASSource.h"
#include "gn/gnGBKSource.h"
#include "gn/gnBaseHeader.h"
#include "gn/gnFilter.h"
#include "gn/gnDebug.h"
#include "gn/gnStringTools.h"
#include <string>

gnDNXSource::gnDNXSource()
{
	m_DNXSpec = new gnGenomeSpec();
	m_pFilter = gnFilter::fullDNASeqFilter();
	if(m_pFilter == NULL){
		DebugMsg("Error using static sequence filters.");
	}
}

gnDNXSource::gnDNXSource( const gnDNXSource& s ) : gnFileSource(s)
{
	if(s.m_DNXSpec != NULL)
		m_DNXSpec = s.m_DNXSpec->Clone();
}

gnDNXSource::~gnDNXSource()
{
	m_ifstream.close();
	delete m_DNXSpec;
}
boolean gnDNXSource::HasContig( const string& name ) const
{
	for(uint32 contigI = 0; contigI < m_DNXSpec->GetSpecListLength(); contigI++){
		if(m_DNXSpec->GetSpec(contigI)->GetName() == name)
			return true;
	}
	return false;
}
uint32 gnDNXSource::GetContigID( const string& name ) const
{
	for(uint32 contigI = 0; contigI < m_DNXSpec->GetSpecListLength(); contigI++){
		if(m_DNXSpec->GetSpec(contigI)->GetName() == name)
			return contigI;
	}
	return ALL_CONTIGS;
}
string gnDNXSource::GetContigName( const uint32 i ) const
{
	if(i < m_DNXSpec->GetSpecListLength()){
		gnBaseSpec *gnbs = m_DNXSpec->GetSpec(i);
		return gnbs->GetName();
	}
	return "";
}
gnSeqI gnDNXSource::GetContigSeqLength( const uint32 i ) const
{
	if( i == ALL_CONTIGS){
		return m_DNXSpec->GetLength();
	}else if(i < m_DNXSpec->GetSpecListLength()){
		gnBaseSpec *gnbs = m_DNXSpec->GetSpec(i);
		return gnbs->GetLength();
	}
	return 0;
}
//read raw data from the file

void gnDNXSource::ValidateName(string& name){
	if(name == ""){	//make a random one.
		name.resize(4);
		srand(time(NULL));
		for(int i=0; i < 4; i++)	
			name[i] = (rand() % 26) + 64;
	}
}

boolean gnDNXSource::Write(gnGenomeSpec* spec, const string& filename){
	ofstream m_ofstream(filename.c_str(), ios::out | ios::binary);
	gnSourceFactory* m_sSourceFactory = gnSourceFactory::GetSourceFactory();
	if(!m_ofstream.is_open())
		return false;
	for(uint32 i=0; i < spec->GetSpecListLength(); i++){	//each of these will be dnx statements
		gnFragmentSpec* curStatementSpec = spec->GetSpec(i);
		string sourceName = spec->GetSourceName();
		string statementName = spec->GetName();
		if(!m_sSourceFactory->HasSource(sourceName)){
			ValidateName(statementName);
			statementName += ".seq";
			m_ofstream << statementName << "=";
		}else
			m_ofstream << sourceName << "=";
		for(uint32 j=0; j < curStatementSpec->GetSpecListLength(); j++){	//each of these will be the files
															//referred to by the dnx statement
			gnContigSpec* curSubSpec = curStatementSpec->GetSpec(i);
			sourceName = curStatementSpec->GetSourceName();
			string contigName = curStatementSpec->GetName();
			if(!m_sSourceFactory->HasSource(sourceName)){
				ValidateName(contigName);
				string writename = contigName+".seq";
				gnSequence gns = *curSubSpec;
				gnGBKSource::Write(gns, writename);
				m_ofstream << writename;
			}else
				m_ofstream << sourceName;
			if( j + 1 < curStatementSpec->GetSpecListLength())
				m_ofstream << "+";
		}
		m_ofstream << ";";
		gnBaseHeader *gpbh = spec->GetHeader(0);
		string header = "";
		if(gpbh != NULL){
			header = gpbh->GetHeader();
			//delete everything after the first newline.
			uint32 newlinepos = header.find_first_of('\n', 0);
			if(newlinepos != string::npos)
				header = header.substr(0, newlinepos - 1);
		}
		m_ofstream << header << "\r\n";
	}
	m_ofstream.close();
	return true;
}

gnFileContig* gnDNXSource::GetFileContig( const uint32 contigI ) const{
	return NULL;	//returning NULL
}

//reads an inputstream and creates fills the spec vector appropriately
boolean gnDNXSource::ParseStream( istream& fin )
{
	// INIT temp varables
	uint32 readState = 0;  //10 - currently inside a comment
	uint32 sectionStart = 0;
	gnFragmentSpec* currentFragSpec = 0;
	gnBaseSource *currentSource;
	string currentSourceName;
	uint32 currentContig = ALL_CONTIGS;
	uint32 currentSeqStart = 0;
	boolean currentRevComp = false;
	// INIT buffer
	uint64 bufReadLen = 0;
	uint64 remainingBuffer = 0;
	Array<char> array_buf( BUFFER_SIZE );
	char* buf = array_buf.data;
	string curliteral;
	
	//Get the source factory and add the current dnx path to it.
	gnSourceFactory *sourceFactory = gnSourceFactory::GetSourceFactory();
	sourceFactory->AddPath(getPathString(m_openString));

	while( !fin.eof() )
	{
		if(sectionStart > 0){
			remainingBuffer = bufReadLen - sectionStart;
			if(readState == 5){	//add literal
				curliteral += string(buf, sectionStart, remainingBuffer);
				remainingBuffer = 0;
				sectionStart = bufReadLen;
			}else
				memmove(buf, buf+sectionStart, remainingBuffer);
		}
		  // read chars
		fin.read( buf + remainingBuffer, BUFFER_SIZE - (bufReadLen - sectionStart));
		sectionStart = 0;
		bufReadLen = fin.gcount() + remainingBuffer;
		
		for( uint32 i=0 ; i < bufReadLen ; i++ )
		{
			char ch = buf[i];
			switch( readState )
			{
				case 0: // Get name of genome
					if(ch == '='){
						//genome name is from sectionStart to i
						string contigName(buf+sectionStart, i - sectionStart);
						currentFragSpec = new gnFragmentSpec();
						currentFragSpec->SetName(contigName);
						currentFragSpec->SetSourceName(m_openString);
						m_DNXSpec->AddSpec(currentFragSpec);
						sectionStart = i+1;
						readState = 1;
					}
					break;
				case 1: // Ignore whitespace before filename or literal
					if((ch == ' ')||(ch == '	'))
						break;
				case 2: // Are we getting a new source file name or a literal?
					if(ch == '"'){ //getting a literal
						readState = 5;
						sectionStart = i+1;
						break;
					}
					readState = 3;
					sectionStart = i;
				case 3: // Get a new source file name
					//stop on >, (, +, and \n
					if(ch == '\n' && sectionStart == i -1){
						if(buf[sectionStart]=='\r'){
							sectionStart = i + 1;
							break;
						}
					}
					if((ch == '+')||(ch == '>')||(ch == '(')||(ch == '\n')||(ch == ';')){
						//use the entire source file
						string seqfile(buf, sectionStart, i - sectionStart);
						currentSourceName = seqfile;
						currentSource = sourceFactory->AddSource(seqfile, true);
						if (currentSource==NULL)
						{
							return false;
						}
						if((ch == '+')||(ch == '\n')||(ch == ';')){
							gnSourceSpec* tmp_spec = new gnSourceSpec(currentSource);
							tmp_spec->SetSourceName(seqfile);
							currentFragSpec->AddSpec(tmp_spec);
							readState = 1;
							if(ch == '\n'){ //reached the end of the statement.  parse another.
								readState = 0;
							}else if(ch == ';'){	//hit a comment.
								readState = 9;
							}
						}else if(ch == '>'){  //select a contig to use
							readState = 4;
						}else if(ch == '('){  // use a specified section of the entire file
							readState = 6;
						}
						sectionStart = i + 1;
					}
					break;
				case 4: // Get a specific contig to use
					//stop on (, +, ;, and \n
					if((ch == '+')||(ch == '\n')||(ch == ';')||(ch == '(')){
						//use the entire contig
						string contigname(buf, sectionStart, i - sectionStart);
						currentContig = currentSource->GetContigID(contigname);
						if((ch == '+')||(ch == '\n')||(ch == ';')){
							gnSourceSpec* tmp_spec = new gnSourceSpec(currentSource, currentContig);
							tmp_spec->SetSourceName(currentSourceName);
							currentFragSpec->AddSpec(tmp_spec);
							readState = 1;
							if(ch == '\n'){ //reached the end of the statement.  parse another.
								readState = 0;
							}else if(ch == ';'){	//hit a comment.
								readState = 9;
							}
						}else if(ch == '('){  //use the specified section
							readState = 6;
						}
						sectionStart = i + 1;
					}
					break;
				case 5: // read in a literal
					// stop on "
					if(ch == '"'){
						//now create a string spec from sectionStart to i-1
						string literal(buf, sectionStart, i - sectionStart);
						if(curliteral.length() > 0){
							literal += curliteral;
							curliteral = "";
						}
						gnStringSpec *gpss = new gnStringSpec(literal, currentFragSpec->GetSpecListLength());
						currentFragSpec->AddSpec(gpss);
					}
				case 6: // read in a specified section
					//stop on , or < or >
					if((ch == ',') || (ch == '<') || (ch == '>')){
						string seqstartstring(buf, sectionStart, i - sectionStart);
						if(seqstartstring == "lend"){
							currentSeqStart = 0;
						}else if (seqstartstring == "rend"){
							currentSeqStart = GNSEQI_END;
						}else
							currentSeqStart = atoi(seqstartstring.c_str()) - 1;
						if(ch == '<')
							currentRevComp = true;
						sectionStart = i + 1;
						readState = 7;
					}
					break;
				case 7: // read in the second half of a specified section
					//stop on )
					if(ch == ')'){
						string seqendstring(buf, sectionStart, i - sectionStart);
						uint32 currentSeqEnd = GNSEQI_END;
						if(seqendstring == "lend"){
							currentSeqEnd = 0;
						}else if (seqendstring == "rend"){
							currentSeqEnd = GNSEQI_END;
						}else
							currentSeqEnd = atoi(seqendstring.c_str()) - 1;
						gnSourceSpec* tmp_spec = new gnSourceSpec(currentSource, currentContig, currentSeqStart, currentSeqEnd, currentRevComp);
						tmp_spec->SetSourceName(currentSourceName);
						currentFragSpec->AddSpec(tmp_spec);
						currentRevComp = false; //set it back to its default value.
						sectionStart = i + 1;
						readState = 8; //look for connective operator
					}
					break;
				case 8: //skip whitespace until a connective or terminating operator is reached.
					if(ch == '+'){
						sectionStart = i + 1;
						readState = 1; //start over
					}
					if(ch == '\n'){
						sectionStart = i + 1;
						readState = 0;
					}
					if(ch == ';'){
						sectionStart = i + 1;
						readState = 9;
					}
					break;
				case 9: //skip comment until newline.
					if(ch == '\n'){
						sectionStart = i + 1;
						readState = 0;
					}
					break;
				default:
					DebugMsg("ERROR in file\n");
					return false;
					break;
			}
		}// for all buf
	}// while !eof
	// CLEAN UP
	return true;
}
